/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/internal.h: Data types, constants, and utility
 * functions exported for use by other OpenGL source files.  These should
 * not be used externally.
 */

#ifndef SIL_SRC_SYSDEP_OPENGL_INTERNAL_H
#define SIL_SRC_SYSDEP_OPENGL_INTERNAL_H

#include "src/math.h"
#include "src/sysdep/opengl/gl-headers.h"

/*************************************************************************/
/************************ Internal-use constants *************************/
/*************************************************************************/

/* Array expansion increment for the delayed-delete buffer. */
#define OPENGL_DELETE_INFO_EXPAND  100

/*************************************************************************/
/****************** Data structures used by OpenGL code ******************/
/*************************************************************************/

/* Framebuffer data structure. */

struct SysFramebuffer {

    /* Linked list pointers for the global framebuffer list. */
    SysFramebuffer *next, *prev;

    /* OpenGL framebuffer ID for this framebuffer. */
    GLuint framebuffer;

    /* OpenGL depth buffer (renderbuffer) ID for this framebuffer. */
    GLuint depth_buffer;

    /* OpenGL stencil buffer (renderbuffer) ID for this framebuffer.  Only
     * valid if separate_stencil is true. */
    GLuint stencil_buffer;

    /* OpenGL texture ID for applying the framebuffer as a texture. */
    GLuint texture;

    /* Size of this framebuffer, in pixels. */
    int width, height;

    /* Depth and stencil buffer formats for this framebuffer. */
    GLenum depth_format;
    GLenum stencil_format;  // Only valid if separate_stencil is true.

    /* Stencil buffer presence flag. */
    uint8_t has_stencil;
    /* Separate depth/stencil buffer flag: true = separate, false = packed.
     * Always false if has_stencil is false. */
    uint8_t separate_stencil;

    /* Antialiasing flag: true = GL_LINEAR, false = GL_NEAREST. */
    uint8_t antialias;

};

/*-----------------------------------------------------------------------*/

/* Graphics primitive data structure. */

struct SysPrimitive {

    /* Linked list pointers for the global primitive list. */
    SysPrimitive *next, *prev;

    /* Primitive type (GL_TRIANGLES, etc.). */
    GLenum type;

    /* Does this primitive have a separate index list? */
    uint8_t has_indices;

    /* Is this primitive using an immediate-mode vertex buffer? */
    uint8_t is_immediate_vbo;

    /* Was this primitive using the shared single-quad index buffer? */
    uint8_t is_single_quad;

    /* Has the vertex and/or index data for this primitive been lost due
     * to device loss or other system error? */
    uint8_t lost;

    /* Are vertex_data and index_data locally-allocated buffers, which
     * should be freed when the primitive is destroyed? */
    uint8_t vertex_local, index_local;

    /* Do vertex_data and index_data contain data temporarily saved from
     * vertex buffers? */
    uint8_t vertex_saved, index_saved;

    /* Has the vertex array object been configured?  (If no vertex array is
     * in use, this is always false.) */
    uint8_t vao_configured;

    /* Vertex buffer object ID, or zero if this primitive has no associated
     * vertex buffer. */
    GLuint vertex_buffer;

    /* Index buffer object ID, or zero if this primitive has no associated
     * index buffer. */
    GLuint index_buffer;

    /* Vertex array object ID, used if the MANDATORY_VAO feature is set. */
    GLuint vertex_array;

    /* Local vertex data buffer.  Always NULL if vertex_buffer != 0. */
    uint8_t *vertex_data;

    /* Vertex data structure size (in bytes per vertex) and count. */
    int vertex_size, vertex_count;

    /* Vertex position information. */
    GLint position_size;        // Number of components (2, 3, or 4).
    GLenum position_type;       // GL_* data type.
    GLsizei position_offset;    // Byte offset from base of array.

    /* Texture coordinate information. */
    GLint texcoord_size;        // 0 if texture coordinates are not used.
    GLenum texcoord_type;
    GLsizei texcoord_offset;

    /* Color information. */
    GLint color_size;
    GLenum color_type;
    GLsizei color_offset;

    /* Local index data buffer.  Always NULL if index_buffer != 0, and
     * also NULL if the primitive has no indices. */
    uint8_t *index_data;

    /* Index data size (in bytes: 1, 2, or 4) and count.  These values are
     * both set to zero if the primitive has no indices. */
    int index_size, index_count;

    /* GL data type of index data. */
    GLenum index_type;

};

/*----------------------------------*/

/* Texture color types. */

enum {
    TEXCOLOR_RGBA = 1,
    TEXCOLOR_RGB,
    TEXCOLOR_A,
    TEXCOLOR_EXTERNAL,  // OpenGL ES only.
};

/* Texture data structure. */

struct SysTexture {

    /* Linked list pointers for the global texture list. */
    SysTexture *next, *prev;

    /* OpenGL texture ID for this texture. */
    GLuint id;

    /* Size of this texture, in pixels. */
    int width, height;

    /* Color type of pixel data (TEXCOLOR_*). */
    uint8_t color_type;

    /* Should we automatically generate mipmaps for this texture? */
    uint8_t auto_mipmaps;

    /* Does this texture currently have mipmaps stored? */
    uint8_t has_mipmaps;

    /* Texture coordinate repeat and antialiasing flags. */
    uint8_t repeat_u, repeat_v;
    uint8_t antialias;

    /* Is the texture empty (newly created and not yet modified)? */
    uint8_t empty;

    /* Is the pixel data readable by attaching the texture to a framebuffer?
     * If false and glGetTexImage() is not available, attempts to lock
     * compressed textures for anything except a complete overwrite
     * (SYS_TEXTURE_LOCK_DISCARD) will fail. */
    uint8_t readable;

    /* Pixel data buffer returned from the last sys_texture_lock(), or
     * NULL if not locked. */
    void *lock_buf;

    /* Lock mode used with the last sys_texture_lock().  Only valid when
     * lock_buf != NULL. */
    SysTextureLockMode lock_mode;

    /* Texture coordinates (in pixels) used with the last sys_texture_lock().
     * Only valid when lock_buf != NULL. */
    int lock_x, lock_y, lock_w, lock_h;

    /* Pixel data buffer for opengl_texture_restore_all(). */
    void *save_buf;

};

/*----------------------------------*/

/* Vertex attribute indices for shaders. */

enum {
    VERTEX_ATTRIB_POSITION = 0,  // attribute vec[234] position;
    VERTEX_ATTRIB_TEXCOORD,      // attribute vec2 texcoord;
    VERTEX_ATTRIB_COLOR,         // attribute vec4 color;
    VERTEX_ATTRIB__NUM           // Total number of defined attributes.
};

/* Uniform indices for shaders (used with opengl_set_uniform_*). */

enum {
    UNIFORM_TRANSFORM = 0,  // uniform mat4 transform;
    UNIFORM_TEXTURE,        // uniform sampler texture;
    UNIFORM_TEX_OFFSET,     // uniform vec2 tex_offset;
    UNIFORM_FIXED_COLOR,    // uniform vec4 fixed_color;
    UNIFORM_FOG_PARAMS,     // uniform vec4 fog_params;
    UNIFORM_FOG_COLOR,      // uniform vec4 fog_color;
    UNIFORM_ALPHA_REF,      // uniform float alpha_ref;
    UNIFORM_POINT_SIZE,     // uniform float point_size;
    UNIFORM__NUM            // Total number of defined uniforms.
};

/*************************************************************************/
/******************* Library-internal shared routines ********************/
/*************************************************************************/

/******** framebuffer.c ********/

/**
 * opengl_current_framebuffer:  Return the framebuffer currently bound as
 * the rendering target, or NULL if the default framebuffer is the current
 * rendering target.
 *
 * [Return value]
 *     Currently bound framebuffer, or NULL if none.
 */
extern SysFramebuffer *opengl_current_framebuffer(void);

/**
 * opengl_framebuffer_clear_all:  Clear all OpenGL resources from all
 * framebuffers.  Called by opengl_save_state().
 */
extern void opengl_framebuffer_clear_all(void);

/**
 * opengl_framebuffer_restore_all:  Recreate OpenGL resources for all
 * framebuffers.  Called by opengl_restore_state().
 */
extern void opengl_framebuffer_restore_all(void);


/******** graphics.c ********/

/**
 * opengl_window_width, opengl_window_height:  Width and height of the
 * OpenGL output window, in pixels.
 */
extern int opengl_window_width, opengl_window_height;

/**
 * opengl_delete_*:  Delete the given GL object.  The type of object is
 * identified by the function name; e.g., opengl_delete_texture() deletes
 * texture objects.  If the DELAYED_DELETE feature is in effect, the object
 * is not deleted immediately, but is instead added to a list for deletion
 * at the next opengl_free_dead_resources() call.
 *
 * [Parameters]
 *     object: GL object to delete.
 */
extern void opengl_delete_buffer(GLuint object);
extern void opengl_delete_framebuffer(GLuint object);
extern void opengl_delete_program(GLuint object);
extern void opengl_delete_renderbuffer(GLuint object);
extern void opengl_delete_shader(GLuint object);
extern void opengl_delete_texture(GLuint object);
extern void opengl_delete_vertex_array(GLuint object);


/******** primitive.c ********/

/**
 * opengl_primitive_reset_bindings:  Reset the vertex and index buffer
 * bindings, ensuring that the cached values are in sync with the GL state.
 * Called at the beginning of each frame to avoid glitches resulting from
 * GL operations performed by system libraries.
 *
 * This routine does nothing if vertex buffers are not enabled (i.e.,
 * neither of the OPENGL_FEATURE_FAST_{STATIC,DYNAMIC}_VBO flags are set).
 */
extern void opengl_primitive_reset_bindings(void);

/**
 * opengl_primitive_save_all:  Save the vertex and index data for all
 * primitives.  Called by opengl_save_state().
 */
extern void opengl_primitive_save_all(void);

/**
 * opengl_texture_restore_all:  Restore the previously-saved vertex and
 * index data for all primitives.  Called by opengl_restore_state().
 */
extern void opengl_primitive_restore_all(void);

/**
 * opengl_primitive_cleanup:  Clear all static buffer objects for
 * immediate-mode primitives.  Called by opengl_save_satte() and
 * opengl_cleanup().
 */
extern void opengl_primitive_cleanup(void);


/******** shader.c ********/

/**
 * opengl_select_shader:  Activate the appropriate shader for the vertex
 * format used by the given primitive, first creating the shader if
 * necessary.  Called by sys_graphics_draw_primitive().
 *
 * [Parameters]
 *     primitive: Primitive from which to take vertex format.
 *     tex_offset: True to apply the texture offset.
 *     color_uniform: True to apply the fixed color parameter.
 *     fog: True to apply fog.
 *     alpha_test: True to perform alpha testing.
 *     alpha_comparison: Alpha test comparison type (GL_GEQUAL, etc.).
 * [Return value]
 *     1 if the active shader was changed, 0 if not, -1 if an error occurred.
 */
extern int opengl_select_shader(const SysPrimitive *primitive, int tex_offset,
                                int color_uniform, int fog, int alpha_test,
                                GLenum alpha_comparison);

/**
 * opengl_set_uniform_*:  Set the value of a uniform parameter to a shader.
 * No type-checking is performed; undefined behavior will result from
 * attempting to load a value of incorrect type into a uniform parameter
 * (for example, trying to load a vector into a matrix).
 *
 * [Parameters]
 *     uniform: ID of uniform parameter to set (UNIFORM_*).
 *     value: Value to set (type depends on function).
 */
extern void opengl_set_uniform_int(int uniform, int value);
extern void opengl_set_uniform_float(int uniform, float value);
extern void opengl_set_uniform_vec2(int uniform, const Vector2f *value);
extern void opengl_set_uniform_vec4(int uniform, const Vector4f *value);
extern void opengl_set_uniform_mat4(int uniform, const Matrix4f *value);

/**
 * opengl_shader_clear_all:  Clear all existing shader objects.  Called
 * by opengl_save_state() and opengl_cleanup().
 */
extern void opengl_shader_clear_all(void);


/******** state.c ********/

/**
 * opengl_framebuffer_changed:  Flag indicating whether the framebuffer set
 * as the render target has been changed.
 */
extern uint8_t opengl_framebuffer_changed;

/**
 * opengl_texture_is_framebuffer:  Flag indicating whether the currently
 * active texture is a framebuffer texture.  Set by sys_framebuffer_apply(),
 * and cleared by sys_texture_apply().
 */
extern uint8_t opengl_texture_is_framebuffer;

/**
 * opengl_current_texture:  Pointer to the currently active texture.  If
 * opengl_texture_is_framebuffer is set, this value is undefined.
 */
extern SysTexture *opengl_current_texture;

/**
 * opengl_current_texture_id:  Current GL_TEXTURE_* target to which a
 * texture is bound, or zero if no texture is bound.
 */
extern GLuint opengl_current_texture_target;

/**
 * opengl_current_texture_id:  Texture ID currently bound to
 * opengl_current_texture_target, or zero if no texture is bound.
 */
extern GLuint opengl_current_texture_id;

/**
 * opengl_primitive_color:  Color parameter to apply to primitives.
 */
extern Vector4f opengl_primitive_color;

/**
 * opengl_primitive_color_used:  Flag indicating whether the current
 * primitive color is not the identity and should therefore be applied.
 * This is merely a shortcut to avoid needing to compare the entire vector
 * against (1,1,1,1).
 */
extern uint8_t opengl_primitive_color_used;

/*----------------------------------*/

/**
 * opengl_state_init:  Initialize internal state data for the OpenGL
 * interface.  Called by opengl_init().
 */
extern void opengl_state_init(void);

/**
 * opengl_state_reapply:  Reapply all internally-stored state values to
 * the OpenGL context.  Called by opengl_restore_state().
 */
extern void opengl_state_reapply(void);

/**
 * opengl_apply_texture:  Bind the given texture to the given target.  Any
 * currently bound texture (on any target) is unbound.  Called by
 * sys_texture_apply() and sys_framebuffer_apply().
 *
 * [Parameters]
 *     target: OpenGL texture target (e.g., GL_TEXTURE_2D).
 *     id: OpenGL texture ID.
 */
extern void opengl_apply_texture(GLenum target, GLuint id);

/**
 * opengl_apply_viewport:  Apply the current viewport parameters.
 * Implements the OpenGL side of sys_graphics_set_viewport(), and is also
 * called on framebuffer change.
 */
extern void opengl_apply_viewport(void);

/**
 * opengl_apply_clip_region:  Apply the current clipping region parameters.
 * Implements the OpenGL side of sys_graphics_set_clip_region(), and is
 * also called on framebuffer change.
 */
extern void opengl_apply_clip_region(void);

/**
 * opengl_apply_matrices:  Apply the current coordinate transformation
 * matrices.  Called by sys_graphics_draw_primitive().
 *
 * [Parameters]
 *     force: True to reapply all matrices regardless of current state;
 *         false to only apply changed matrices.
 */
extern void opengl_apply_matrices(int force);

/**
 * opengl_apply_shader:  Select an appropriate shader for rendering the
 * given primitive based on the current rendering state.  Called by
 * sys_graphics_draw_primitive() if shaders are enabled.
 *
 * [Parameters]
 *     primitive: Primitive to be rendered.
 * [Return value]
 *     True on success, false on error.
 */
extern int opengl_apply_shader(SysPrimitive *primitive);


/******** texture.c ********/

/**
 * opengl_texture_save_all:  Save the pixel data for all textures.  Called
 * by opengl_save_state().
 */
extern void opengl_texture_save_all(void);

/**
 * opengl_texture_restore_all:  Restore the previously-saved pixel data for
 * all textures.  Called by opengl_restore_state().
 */
extern void opengl_texture_restore_all(void);

/*************************************************************************/
/****************** FYI: Caveats for particular systems ******************/
/*************************************************************************/

/*
 * - NVIDIA's OpenGL ES 1.1 library for the Tegra chipset (as used in
 *   Android 4.2, at least), as well as Android's built-in libagl, lack
 *   support for several required glGet() targets, notably
 *   GL_FRAMEBUFFER_BINDING and GL_TEXTURE_BINDING_EXTERNAL_OES.
 *   Similarly, glIsEnabled(GL_TEXTURE_EXTERNAL_OES) fails with
 *   INVALID_ENUM in these environments.
 */

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_OPENGL_INTERNAL_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
